
unit SystemFolderList;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ShellApi, ShlObj, ComObj, Commctrl, ActiveX, StdCtrls, ExtCtrls,
  Menus;


type
  tStateClickEvent = procedure ( StateChecked : boolean ) of object;
  TSystemFolderList = class(TCustomTreeView)
  private
   fDirectory : string;
   fPidlPath : PItemIdList;
   fCheckStates : boolean;
   fCanExpand : boolean;
   fItemsCount : integer;
   fStateClickEvent : tStateClickEvent;
   fListCheckedFolder : tStringList;
  protected
   procedure CreateWnd; override;
   procedure DestroyWnd; override;
   procedure Loaded; override;
   function GetCheckStates : boolean;
   procedure SetCheckStates ( states : boolean );
   function GetStateChecked : boolean;
   procedure SetStateChecked ( StateChecked : boolean );
   function GetListCheckedFolder : tStringList;
   procedure AddCheckedFolder ( folder : string );
   function InitSystemFolderList : TTreeNode;
   function AddFolderToTreeView ( folder : string ) : TTreeNode;
   procedure AddSubFolderToTreeView ( node : tTreeNode );
   procedure EnumerateFolder (  ShellFolder: IShellFolder; mask : DWORD;
                                                            node : tTreeNode );
   procedure ClearAllItems;
   procedure CNNotify ( var Message: tWMNotify ); message CN_NOTIFY;
   function MakeHitTest : tTreeNode;
   function ProofStates ( node : tTreeNode ) : boolean;
   procedure KeyDown (  var Key: Word; Shift: TShiftState ); override;
   function CanExpand ( node: tTreeNode ) : boolean; override;
   function CanCollapse ( node: tTreeNode ) : boolean; override;
   function CheckSibling ( node : tTreeNode ) : boolean;
   procedure CheckParentFolder ( node : tTreeNode );
   procedure CheckSubFolder ( node : tTreeNode );
   procedure CreateImages;
   procedure CreateStates;
   public
   constructor create (aOwner: tComponent); override;
   destructor destroy; override;
  published
   property Directory : string read fDirectory write fDirectory;
   property PidlPath : PItemIdList read fPidlPath;
   property StateBoxes : boolean read GetCheckStates write SetCheckStates
    default false;
   property StateChecked : boolean read GetStateChecked write SetStateChecked
    default false;
   property OnStateClick : tStateClickEvent read fStateClickEvent
                                             write fStateClickEvent;
   property ListCheckedFolder : tStringList read GetListCheckedFolder
                                            write fListCheckedFolder;
   property ReadOnly;
   property Align;
   property Enabled;
   property Font;
   property HideSelection;
   property Items;
   property ParentColor;
   property PopupMenu;
   property Selected;
   property TabOrder;
   property TabStop default true;
   property Visible;
   property OnChanging;
   property OnChange;
   property OnClick;
   property OnMouseDown;
   property OnKeyDown;
  end;


{ *************************** SUBROUTINES ************************************* }
procedure SplitCurrentDir ( FolderList : tStringList );
function AddPIDLs ( id1, id2 : PItemIDList ) : PItemIDList;

function GetPidlSize ( pidl: PItemIdList ) : integer;
function GetFolderName ( ShellFolder : IShellFolder;
                                 idList : PItemIDList; Flags : DWORD ) : string;
function GetSelectedIcon ( list : PItemIdList ) : integer;
function GetNormalIcon ( list : PItemIdList ) : integer;


procedure Register;

implementation

{$R SystemFolderList.res}


const
  NOTCHECKED = 1;
  CHECKED = 2;
  GRAYCHECKED = 3;
  TIMERINTERVALL = 100;
var
  StateImageList : tImageList;
  bitmap : array[1..4] of tBitmap;
  RootFolder : iShellFolder;
  ShellMalloc : iMalloc;
  CurrentNode : tTreeNode;
  FullPidlPathList : tList;
  SomeFolderChecked : boolean;



Procedure SplitCurrentDir ( FolderList : tStringList );
var
  i : integer;
  folder : string;
  CurDir : string;

Begin
  CurDir := GetCurrentDir;
  i := pos ( '\', CurDir  );
  while i > 0 do
   begin
    folder := copy ( CurDir, 1, i-1 );
    FolderList.add ( folder );
    delete ( CurDir, 1, i );
    i := pos ( '\', CurDir  );
   end;
  if length ( CurDir ) > 3 then FolderList.add ( CurDir );
End;


Function AddPidls ( id1, id2 : pItemIdList ) : pItemIdList;
var
  s1, s2 : UINT;
begin
  if ( id1 <> nil ) then
   s1 := GetPidlSize ( id1 )
    else
     s1 := 0;
  s2 := id2.mkid.cb;
  result := ShellMalloc.Alloc ( s1 + s2 + SizeOf(word) );

  if result <> nil then
  begin
   if ( id1 <> nil ) then
    Move ( id1^, result^, s1 );
   Move ( id2^, PChar(result)[s1], s2 + SizeOf(word) );
  end;
end;


Function GetPidlSize ( pidl : pItemIdList ) : integer;
Begin
  result := 0;
  if pidl <> nil then
  begin
   while pidl^.mkid.cb <> 0 do
    begin
     inc ( result, pidl^.mkid.cb );
     inc ( longint(pidl), pidl^.mkid.cb );
    end;
  end;
End;


Function GetFolderName ( ShellFolder : iShellFolder;
                                 idList : pItemIdList; flags : DWORD ) : string;
var
  str : tStrRet;
  name : string;
Begin
  name := '';
  if ShellFolder.GetDisplayNameOf ( idList, flags, str ) = NOERROR then
  begin
    case str.uType of
     STRRET_WSTR :
      name := WideCharToString(str.pOleStr);
     STRRET_OFFSET :
      name := PChar(longint(IdList) + str.uOffset);
     STRRET_CSTR :
      name := str.cStr;
    end;
  end;
  if ( pos ( '\', name ) > 0 ) and ( length(name) = 3 ) then
   SetLength ( name, 2 );
  result := name;
End;


Function GetNormalIcon ( list : pItemIdList ) : integer;
var
  sfi : tShFileInfo;
Begin
  ShGetFileInfo ( PChar(list), 0, sfi, SizeOf(tShFileInfo),
                          SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_SMALLICON );
  result := sfi.iIcon;
End;


Function GetSelectedIcon ( list : pItemIdList ) : integer;
var
  sfi : tShFileInfo;
Begin
  ShGetFileInfo ( PChar(list), 0, sfi, SizeOf(tShFileInfo),
        SHGFI_PIDL or SHGFI_SYSICONINDEX or SHGFI_SMALLICON or SHGFI_OPENICON );
  result := sfi.iIcon;
End;


constructor tSystemFolderList.create;
Begin
  inherited create(aOwner);
  width := 220;
  height := 140;
  CreateStates;
  CreateImages;
  fCheckStates := false;
  fCanExpand := true;
  FullPidlPathList := tList.create;
  fListCheckedFolder := tStringList.create;
End;


destructor tSystemFolderList.destroy;
var
  i : integer;
Begin
  ClearAllItems;
  FullPidlPathList.free;
  fListCheckedFolder.free;
  images.free;
  if assigned ( StateImages ) then
   begin
    StateImageList.free;
    for i := 1 to 4 do bitmap[i].free;
   end;
  inherited;
End;


Procedure tSystemFolderList.CreateWnd;
Begin
  inherited CreateWnd;

  font.size := 8;
  font.name := 'arial';

  if not ( csLoading in ComponentState ) then
   begin
    CurrentNode := InitSystemFolderList;
    CurrentNode.selected := true;
   end;
End;


Procedure tSystemFolderList.DestroyWnd;
Begin
  inherited DestroyWnd;
End;


Procedure tSystemFolderList.Loaded;
Begin
  inherited Loaded;
  CurrentNode := InitSystemFolderList;
  CurrentNode.selected := true;
End;


Function tSystemFolderList.GetCheckStates : boolean;
Begin
  result := fCheckStates;
End;


Procedure tSystemFolderList.SetCheckStates ( states : boolean );
Begin
  if states then
   if not assigned ( StateImages ) then
    begin
     fCheckStates := true;
     StateImages := StateImageList;
    end;
  if not states then
      begin
       fCheckStates := false;
       StateImages := nil;
      end;
End;


Function tSystemFolderList.GetStateChecked : boolean;
var
  node : tTreeNode;
Begin
  result := false;
  node := selected;
  if node <> nil then
   if node.StateIndex = CHECKED  then result := true;
End;


Procedure tSystemFolderList.SetStateChecked ( StateChecked : boolean );
var
  node : tTreeNode;
Begin
  node := selected;
  if node <> nil then
   begin
    if StateChecked = true then node.StateIndex := CHECKED
     else
      if not StateChecked then node.StateIndex := NOTCHECKED;
   end;
End;



Function tSystemFolderList.GetListCheckedFolder : tStringList;
var
  node : tTreeNode;
  path : string;

Begin
//  screen.cursor := crHourGlass;
  fListCheckedFolder.clear;
  node := items[1];
  while node <> nil do
   begin
    fPidlPath := node.data;
    path := GetFolderName ( RootFolder, fPidlPath, SHGDN_FORPARSING );
    if Node.StateIndex = CHECKED then
     begin
      fListCheckedFolder.add ( path );
      AddCheckedFolder ( path );
      if Node.GetNextSibling <> nil
       then Node := Node.GetNextSibling
        else Node := Node.GetNext;
     end
      else Node := Node.GetNext;
   end; // end while do ..
//  screen.cursor := crDefault;
  result := fListCheckedFolder;
End;


Procedure tSystemFolderList.AddCheckedFolder ( folder : string );

var
   result : integer;
   f : tSearchRec;
   SubFolder : string;
   index : integer;

Begin
   result := FindFirst ( folder + '\*.*', faAnyFile, f );
   while result = 0 do
     begin
      if ( f.name[1] <> '.' ) then
       if ( f.attr and faDirectory > 0 ) then
        begin                             { if folder then.. }
         SubFolder := f.name;
         folder := folder + '\' + SubFolder;
         fListCheckedFolder.add ( folder );
         AddCheckedFolder ( folder );
         index := length(folder) - length(SubFolder);
         delete ( folder, index, length(SubFolder) + 1 );
        end;
      result := FindNext( f );
     end; { end while do.. }
   FindClose ( F );
End;


Function tSystemFolderList.InitSystemFolderList : TTreeNode;
var
  AbsList : pItemIDList;
  FolderName : string;
  RootId : pItemIDList;
  ShellFolder : iShellFolder;
  node : tTreeNode;
  i : integer;
  FolderList : tStringList;

Begin
  ClearAllItems;
  FolderList := tStringList.create;
  ShGetDesktopFolder ( RootFolder );
  ShGetSpecialFolderLocation ( handle, CSIDL_DRIVES, RootId );
  AbsList := AddPidls ( nil, RootId );
  FolderName := GetFolderName ( RootFolder, RootId, SHGDN_NORMAL );
  Node := items.AddChildObject ( nil, FolderName, AbsList );
  Node.ImageIndex := GetNormalIcon ( AbsList );
  Node.SelectedIndex := GetSelectedIcon ( AbsList );
  Node.StateIndex := 0;
  // ( Add drives )
  RootFolder.BindToObject ( AbsList, nil, IID_IShellFolder, pointer(ShellFolder));
  EnumerateFolder ( ShellFolder, SFGAO_FILESYSTEM, node );
  // Split Current-Path in seperate folder.
  SplitCurrentDir ( FolderList );
  // ( Build the Current-Path in the TreeView. )
  for i := 0 to FolderList.count-1 do
   node := AddFolderToTreeView ( FolderList[i] );
  // Make Current-Path visible.
  if node <> nil then node.MakeVisible;
  fItemsCount := items.count;
  result := node;
End;


Function tSystemFolderList.AddFolderToTreeView ( folder : string ) : TTreeNode;

var
  PidlPath : pItemIDList;
  ShellFolder : iShellFolder;
  node : tTreeNode;
  i : integer;
  drive : string;

Begin
  // Search current folder in TreeView.
  node := items[0];
  for i := 0 to  items.count-1 do
   begin
    drive := node.text;
    if pos ( folder, drive ) > 0 then break;
    node := node.GetNext;
   end;
  node := items[i];
  PidlPath := node.data;
  // Get System-Folder in TreeView. 
  RootFolder.BindToObject ( PidlPath, nil, IID_IShellFolder, pointer(ShellFolder));
  EnumerateFolder ( ShellFolder, SFGAO_FILESYSTEM, node );
  //TreeView_SortChildren ( handle, node.ItemId, 0 );
  result := node;
End;


Procedure tSystemFolderList.EnumerateFolder ( ShellFolder: IShellFolder;
                                               mask : DWORD; node : tTreeNode );
var
  flags : DWORD;
  EnumList : IEnumIdList;
  PidlPathList, PidlPath, list : PItemIdList;
  fetched : longint;
  FolderName: string;
  attr : integer;

Begin
  PidlPath := node.data;
  flags := SHCONTF_FOLDERS or SHCONTF_INCLUDEHIDDEN;
  if ( ShellFolder.EnumObjects ( handle, flags, EnumList )) = NOERROR then
   begin
    while EnumList.Next ( 1, list, fetched ) = S_OK do
     begin
      FolderName := GetFolderName ( ShellFolder, list, SHGDN_NORMAL );
      attr := SFGAO_DISPLAYATTRMASK or SFGAO_CAPABILITYMASK or $F0000000;
      ShellFolder.GetAttributesOf ( 1, list, attr );

      if ( attr and mask <> 0 ) then
       begin
        PidlPathList := AddPidls ( PidlPath, list );
        FullPidlPathList.add ( PidlPathList );
        node := items.AddChildObject ( node, FolderName, PidlPathList );
        if ( Attr and SFGAO_HASSUBFOLDER <> 0 ) then
         node.HasChildren := true;
        node.ImageIndex := GetNormalIcon ( PidlPathList );
        node.SelectedIndex := GetSelectedIcon ( PidlPathList );
        node.StateIndex := NOTCHECKED;
        if node.parent.StateIndex = CHECKED then
         node.StateIndex := CHECKED
          else node.StateIndex := NOTCHECKED;
        node := node.parent;
       end;
      if assigned ( list )
       then ShellMalloc.free ( list );
     end; // end while do ..
   end;
End;


Procedure tSystemFolderList.CNNotify ( var Message : tWMNotify );
var
  node : tTreeNode;
  StateChecked : boolean;

Begin
  case  message.nmhdr.code of
   TVN_SELCHANGED :
    begin
     node := selected;
     fPidlPath := node.data;
     if fPidlPath <> nil then
      fDirectory := GetFolderName ( RootFolder, fPidlPath, SHGDN_FORPARSING );
    end;
   NM_CLICK : // if left Mouse-Key pressed.
    begin
     fCanExpand := true;
     node := MakeHitTest;
     if node <> nil then
      begin
       if fCanExpand then AddSubFolderToTreeView ( node )
        else
         if fCheckStates then if not fCanExpand then
          begin
           StateChecked := ProofStates ( node );
           if assigned (fStateClickEvent) then fStateClickEvent ( StateChecked );
          end;
      end;
    end;
  end; // case end..
  inherited;
End;


Procedure tSystemFolderList.KeyDown (  var Key : Word; Shift: TShiftState );
var
  node : tTreeNode;
Begin
  fCanExpand := true;
  node := selected;
  // if right Cursor-Key is pressed.
  if Key = VK_RIGHT then AddSubFolderToTreeView ( node );
  inherited;
End;


Function tSystemFolderList.CanExpand ( node: TTreeNode ) : boolean;
Begin
  result := fCanExpand;
End;


Function tSystemFolderList.CanCollapse ( node: TTreeNode ) : boolean;
Begin
  result := fCanExpand;
End;


Procedure tSystemFolderList.AddSubFolderToTreeView ( node : tTreeNode );
var
  idList : pItemIDList;
  ShellFolder : iShellFolder;

Begin
  if ( node.HasChildren ) and ( node.GetFirstChild = nil ) then
   begin
    screen.cursor := crHourGlass;
    idList := node.data;
    RootFolder.BindToObject ( idList,  nil, IID_IShellFolder,
                                                       pointer ( ShellFolder ));
    EnumerateFolder ( ShellFolder, SFGAO_FOLDER, node );
    // EnumerateFolder ( ShellFolder, SFGAO_FILESYSTEM, node );
    TreeView_SortChildren ( handle, node.ItemId, 0 ); // hTreeItem
    Screen.Cursor := crDefault;
    fItemsCount := items.count;
   end;
End;


Function tSystemFolderList.MakeHitTest : tTreeNode;
var
  tvhti : tTVHitTestInfo;
  pos : DWORD;

Begin
  fCanExpand := true;
  pos := GetMessagePos();
  tvhti.pt.x := LOWORD ( pos );
  tvhti.pt.y := HIWORD ( pos );

  MapWindowPoints ( HWND_DESKTOP, handle, tvhti.pt, 1 );
  TreeView_HitTest ( handle, tvhti );
  if ( tvhti.flags and TVHT_ONITEMSTATEICON ) > 0 then
   fCanExpand := false
    else if ( tvhti.flags and TVHT_ONITEMICON ) > 0 then
     fCanExpand := true;
  result := GetNodeAt ( tvhti.pt.x, tvhti.pt.y );
End;


Function tSystemFolderList.ProofStates ( node : tTreeNode ) : boolean;
var
  rect : tRect;

Begin
  rect := ClientRect;
  fCanExpand := false;
  result := false;
  if ( Node.StateIndex = NOTCHECKED ) or ( Node.StateIndex = GRAYCHECKED ) then
   begin
    node.StateIndex := CHECKED;
    result := true;
   end
    else if ( node.StateIndex = CHECKED ) then Node.StateIndex := NOTCHECKED;
  CheckSubFolder ( node );
  CheckParentFolder ( node );
  ValidateRect ( handle, @rect );         { because of blinking. }
  InvalidateRect ( handle, @rect, false );
  UpdateWindow ( handle );
End;


Procedure tSystemFolderList.CheckSubFolder ( node : tTreeNode );
var
  state : integer;
  level : integer;
  folder : string;

Begin
  level := node.level;
  state := node.StateIndex;
  node := node.GetFirstChild;
  while node <> nil do
   begin
    if ( node = nil ) or ( node.level <= level ) then
     break;
    folder := node.text;
    Node.StateIndex := state;
    Node := Node.GetNext;
   end;
End;


Procedure tSystemFolderList.CheckParentFolder ( node : tTreeNode );
var
  AllFolderChecked : boolean;
  folder : string;

Begin
  while node.parent.level > 0 do
   begin
    folder := node.parent.text;
    AllFolderChecked := CheckSibling ( node.parent );

    if AllFolderChecked then
     Node.Parent.StateIndex := CHECKED
     else
      if ( not AllFolderChecked ) and ( not SomeFolderChecked ) then
       Node.Parent.StateIndex := NOTCHECKED
       else
        if SomeFolderChecked then
         Node.Parent.StateIndex := GRAYCHECKED;
    node := node.parent;
   end;
End;


Function tSystemFolderList.CheckSibling ( node : tTreeNode ) : boolean;
var
  folder : string;

Begin
  result := true; // all folder are checked.
  SomeFolderChecked := false;

  Node := Node.GetFirstChild;
  while node <> nil do
   begin
    folder := node.text;
    if ( Node.StateIndex = CHECKED ) or ( Node.StateIndex = GRAYCHECKED ) then
     SomeFolderChecked := true
      else
       if Node.StateIndex = NOTCHECKED then
        result := false;
    Node := Node.GetNextSibling;
   end;
End;


Procedure tSystemFolderList.CreateStates;
var
  i : integer;

Begin
  StateImageList := tImageList.create(self);
  for i := 1 to 4 do bitmap[i] := tBitmap.create;
  bitmap[1].handle := LoadBitmap ( hInstance, 'TV_STATUS1'  );
  { I know not, why the first bitmap do not work ? }
  bitmap[2].handle := LoadBitmap ( hInstance, 'TV_STATUS2'  );
  bitmap[3].handle := LoadBitmap ( hInstance, 'TV_STATUS3' );
  bitmap[4].handle := LoadBitmap ( hInstance, 'TV_STATUS4' );
  for i := 1 to 4 do StateImageList.add ( bitmap[i], nil );
End;


Procedure tSystemFolderList.CreateImages;
var
 SysIL : uint;
 sfi : tshFileInfo;

Begin
  Images := tImageList.Create( self );
  SysIL := ShGetFileInfo ( '', 0, sfi, SizeOf(sfi), SHGFI_SYSICONINDEX or
                                                        SHGFI_SMALLICON );
  if SysIL <> 0 then
   begin
    Images.Handle := SysIL;
    Images.ShareImages := true;
   end;
End;


Procedure tSystemFolderList.ClearAllItems;
var
  i : integer;

Begin
  for i := 0 to FullPidlPathList.count-1 do
  begin
   if FullPidlPathList[i] <> nil then
    ShellMalloc.free ( FullPidlPathList[i] );
  end;
  items.clear;
End;


Procedure Register;
Begin
  RegisterComponents ( 'Test', [TSystemFolderList] );
End;


initialization
  ShGetMalloc ( ShellMalloc );


END.
